name = 'fars_cleaner'
__version__ = "1.2.0"

from fars_cleaner.datasets import FARSFetcher
from fars_cleaner.fars_processor import FARSProcessor
from fars_cleaner.data_loader import load_pipeline
#from .data_loader import pipeline
